/*=============================================================================
	This is not supposed to build and should only be used for learning purposes
=============================================================================*/

//DO NOT CALL IF THERE'S NO ENTRY LEVEL
void UXC_GameEngine::CleanupTransientRefs()
{
	guard( UXC_GameEngine::CleanupTransientRefs);
	FTransientClassMap ClassMap(0);
	ULevel* EntryLevel = GetEntryLevel(this);
	ULevel* ActiveLevel = GetActiveLevel(this);

	if ( !ActiveLevel || (ActiveLevel == EntryLevel) )
		return;

	INT ScriptedTexIdx = ClassMap.RegisterClass( UScriptedTexture::StaticClass() );
	INT Idx = 0;

	for ( TObjectIterator<UObject> It; It; ++It )
	{
		if ( It->GetOuter() )
		{
			if ( It->IsA(UScriptedTexture::StaticClass()) )
				Idx = ScriptedTexIdx;
			else
			{
				Idx = 0;
				UObject* TheOuter = It->GetOuter();
				while ( TheOuter && (TheOuter->GetOuter() != TheOuter) )
				{
					if ( (TheOuter == GetTransientPackage()) || (EntryLevel && (TheOuter == EntryLevel)) )
					{
						Idx = ClassMap.RegisterClass( It->GetClass() );
						break;
					}
					TheOuter = TheOuter->GetOuter();
				}
			}
			if ( ClassMap(Idx).TotalProps <= 0 )
				continue;
//			if ( Idx > 0 )
//				debugf( NAME_Log, TEXT("[XCGE]--- Scanning object %s"), It->GetName() );
			while ( Idx > 0 )
			{
				for ( int i=0 ; i<ClassMap(Idx).Properties.Num() ; i++ )
				{
					UProperty* Prop = ClassMap(Idx).Properties(i);
//					debugf( NAME_Log, TEXT("[XCGE]---- Scanning property %s"), Prop->GetName() );
					check( Prop);
					UObject* TheObj = *It;
					AActor** Ref = (AActor**) ( DWORD (TheObj) + Prop->Offset);
					for ( int j=0 ; j<Prop->ArrayDim ; j++ )
					{
						if ( *Ref && ((*Ref)->GetOuter() == ActiveLevel->GetOuter() ) )
						{
							if ( bEnableDebugLogs )
								debugf( (EName)XC_ENGINE_XC_Engine.GetIndex(), TEXT("[WARNING] Dangerous actor reference held by %s: %s[%i]=%s, nulling out."), It->GetFullName(), Prop->GetName(), j, (*Ref)->GetFullName() );
							*Ref = NULL;
						}
						Ref++;
					}
				}
				Idx = ClassMap(Idx).ParentIdx;
			}
		}
	}

	unguard;
}


INT FTransientClassMap::RegisterClass( UClass* InClass) //Return index in map
{
	guard(FTransientClassMap::RegisterClass);
	if ( !ArrayNum )
	{
		AddZeroed();
		(*this)(0).ParentIdx = -1;
		(*this)(0).Class = UObject::StaticClass();
		if ( InClass == UObject::StaticClass() )
			return 0;
	}
	//Find direct subclasses, register whatever's necessary
	UClass* Parent = UObject::StaticClass();
	INT ParentIdx = 0;
	while ( Parent )
	{
		UBOOL Found = 0;
		UClass* CurClass = InClass;
		if ( CurClass == Parent ) //Cur has been added!
			break;
		while ( CurClass->GetSuperClass() != Parent )
			CurClass = CurClass->GetSuperClass();
		//This class's parent is obviously on this list, attempt to find using parent's children array
		for ( int i=0 ; i< (*this)(ParentIdx).ChildrenIdx.Num() ; i++ )
		{
			if ( (*this)( (*this)(ParentIdx).ChildrenIdx(i) ).Class == CurClass ) //This class is listed on parent's children, do not register
			{
				Found = 1;
				ParentIdx = (*this)(ParentIdx).ChildrenIdx(i);
				break;
			}
		}
		//Register this class if necessary
		if ( !Found )
		{
			AddZeroed();
			Last().Class = CurClass;
			Last().Idx = ArrayNum - 1;
			Last().ParentIdx = ParentIdx;
			Last().TotalProps = (*this)(ParentIdx).TotalProps;
			(*this)(ParentIdx).ChildrenIdx.Add();
			(*this)(ParentIdx).ChildrenIdx.Last() = Last().Idx; //Add the new child IDX (this) onto the parent
			
//			debugf( NAME_Log, TEXT("[XCGE] Registering class %s"), CurClass->GetName() );

			UField* Field = CurClass->Children;
			//Exceptions!
			if ( !CurClass->IsChildOf(UPlayer::StaticClass()) && !CurClass->IsChildOf(USubsystem::StaticClass()) )
			{
				while ( Field ) //Find Actor references
				{
					if ( Field->IsA( UObjectProperty::StaticClass()) && !(((UProperty*)Field)->PropertyFlags & CPF_Native) && ((UObjectProperty*)Field)->PropertyClass->IsChildOf(AActor::StaticClass() ) ) //Do not list native props
					{
//						debugf( NAME_Log, TEXT("[XCGE]-- Registering actor property %s"), Field->GetName() );
						Last().Properties.AddZeroed();
						Last().Properties.Last() = (UProperty*) Field;
						Last().TotalProps++;
					}
					Field = Field->Next;
				}
			}
			ParentIdx = Last().Idx;
		}
		Parent = CurClass;
	}
	return ParentIdx;
	unguard;
}

//==============================================================================
// FCollisionCacus experiment
//==============================================================================
void UXC_GameEngine::SetupCollisionCacus( ULevel* Lev)
{
	guard( UXC_GameEngine::SetupCollisionCacus);
	if ( !bXC_Init || !Lev || !Lev->Hash || (Lev->Hash == CollisionCacus) )
		return;
	if ( !CollisionCacus )
	{
		if ( Lev->GetOuter() )
			debugf( (EName)XC_ENGINE_XC_Engine.GetIndex(), TEXT("Setting up collision octree for %s"), Lev->GetOuter()->GetName() );

		CollisionCacus = new FCollisionCacus;
		guard( LevelBox);
		//THIS WILL SUCK ON 227'S ADDITIVE LEVELS!
		if ( Lev->Model->Points.Num() )
		{
			FBox LBox = FBox( &Lev->Model->Points(0), Lev->Model->Points.Num() );
			LBox.Min -= FVector( 64, 64, 64);
			LBox.Max += FVector( 64, 64, 64);
			CollisionCacus->BoundingBox = LBox;
			CollisionCacus->CenterPoint = (LBox.Max+LBox.Min) * 0.5;
			if ( bEnableDebugLogs )
				debugf( (EName)XC_ENGINE_XC_Engine.GetIndex(), TEXT("Altered octree box to (%i,%i,%i)(%i,%i,%i)"), (INT)LBox.Min.X, (INT)LBox.Min.Y, (INT)LBox.Min.Z, (INT)LBox.Max.X, (INT)LBox.Max.Y, (INT)LBox.Max.Z );
		}
		unguard;
		CollisionCacus->SetupGrids();
		guard( AddActors);
		for( INT i=0 ; i<Lev->Actors.Num() ; i++ )
			if( Lev->Actors(i) && !Lev->Actors(i)->bDeleteMe && Lev->Actors(i)->bCollideActors )
			{
				Lev->Hash->RemoveActor( Lev->Actors(i) );
				CollisionCacus->AddActor( Lev->Actors(i) );
			}
		unguard;
	}
	CollisionCacus->Hooked = Lev->Hash;
	Lev->Hash = CollisionCacus;
	unguard;
}

void UXC_GameEngine::DeleteCollisionCacus( ULevel* Lev)
{
	guard( UXC_GameEngine::DeleteCollisionCacus);
	if ( !CollisionCacus )
		return;
	if ( Lev && CollisionCacus->Hooked )
		Lev->Hash = CollisionCacus->Hooked;
	delete CollisionCacus;
	CollisionCacus = NULL;
	unguard;
}


//==============================================================================
// Dummy Pawn functions
//==============================================================================

//Called mostly on Actor context
void AXC_DummyPawn::NewPlayerCanSeeMe(FFrame &Stack, RESULT_DECL)
{
	guard( XC_Engine::NewPlayerCanSeeMe);
	P_FINISH;

	*(UBOOL*)Result = 0;
	if ( XLevel )
	{
		if ( XLevel->Engine->Client && XLevel->Engine->Client->Viewports.Num() && XLevel->Engine->Client->Viewports(0) && XLevel->Engine->Client->Viewports(0)->Actor ) //Local player
			if ( PCanSee( XLevel->Engine->Client->Viewports(0)->Actor) )
			{
				*(UBOOL*)Result = 1;
				return;
			}
		if ( XLevel->NetDriver && XLevel->NetDriver->ClientConnections.Num() ) //Clients
		{
			for ( INT ip=0 ; ip < XLevel->NetDriver->ClientConnections.Num() ; ip++ )
				if ( XLevel->NetDriver->ClientConnections(ip) && XLevel->NetDriver->ClientConnections(ip)->Actor )
					if ( PCanSee( XLevel->NetDriver->ClientConnections(ip)->Actor) )
					{
						*(UBOOL*)Result = 1;
						return;
					}
		}
	}

	unguard;
}

UBOOL AXC_DummyPawn::PCanSee( APlayerPawn* Other)
{
	guard( XC_Engine::PCanSee);
	check( Other);
	FVector aVec = Location - Other->Location;
	if ( aVec.IsZero() )
		return 0;
	if ( aVec.SizeSquared() > 2560000 )
		return 0;
	if ( aVec.SizeSquared() > (aVec + Other->ViewRotation.Vector()).SizeSquared() )
		return 0;
	if ( !aVec.Normalize() )
		return 0;
	FLOAT aFov = Other->FovAngle / 114.6; //Real angle outside of the center
	FLOAT aDot = aVec|Other->ViewRotation.Vector(); //This is the cosine of our angle
	if ( aDot < appCos(aFov * 1.2) )
		return 0;
	FCheckResult Hit;
	return XLevel->SingleLineCheck( Hit, this, Location, Other->Location + FVector(0,0,Other->BaseEyeHeight), TRACE_Level, FVector(0,0,0), 0|NF_NotVisBlocking);
	unguard;
}